#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/types.h>
#include <regex.h>

#define BUFSIZE 300
#include "Inputtools2.h"

#undef DEBUG


#define cCOMMENT1  '!'
#define cCOMMENT2  '#'
  
static char *delim=" ";

static int strlen_trim(const char *str)
{
  int len;
  int i;
  len = strlen(str);
  for (i=len-1;i>=0;i--) {
    if (str[i]!=' ' && str[i]!='\t' && str[i]!='\n') return i+1;
  }
  return 0;
}

static char *string_adjustl(char *s)
{
  char *c,*dest,*src;

  for (c=s; *c!='\0'; c++) {
    if (*c==' '|| *c=='\t' || *c=='\n') {continue;}
    else { break; }
  }
  if (*c=='\0') return s;
  if (c==s) return s;
  for (src=c,dest=s; *src!='\0'; *dest++=*src++) ;
  *dest='\0';
  return s;
}

static char *string_trim(char *s)
{
  int n;
  n=strlen_trim(s);
  if (n==0) return s;
  s[n]='\0';
  return s;
}


static int is_integer(char *string)
{
#ifdef CHECK_TYPE
  regex_t preg; 
  size_t nmatch=10; 
  regmatch_t pmatch[10];
  int eflags;  
  int ret;
  int len2;
  int i;

  regcomp(&preg, "[+-]?[0-9]+", REG_ICASE|  REG_EXTENDED);

  ret= regexec(  &preg,  string,
                 nmatch,   pmatch, 0); 
  if (ret==0) { /*match*/
    i=0;
    len2=pmatch[i].rm_eo-pmatch[i].rm_so; 
    if (strlen(string)==len2) {
      regfree(&preg);
      return 1; /* the same size  */
    }
  }
  regfree(&preg); /* free memory */
  return 0;
#else
  return 1;
#endif
}

static int is_double(char *string)
{
#ifdef CHECK_TYPE
  regex_t preg;
  size_t nmatch=10;
  regmatch_t pmatch[10];
  int eflags;
  int ret;
  int len2;
  int i;

  regcomp(&preg, "[+-]?[0-9]*[.]?[0-9]*([de][+-]*[0-9]+)?", REG_ICASE|  REG_EXTENDED);

  ret= regexec(  &preg,  string,
                 nmatch,   pmatch, 0);
  if (ret==0) { /*match*/
    i=0;
    len2=pmatch[i].rm_eo-pmatch[i].rm_so;
    if (strlen(string)==len2) {
      regfree(&preg);
      return 1;
    }
  }
  regfree(&preg);
  return 0;

#else
  return 1;
#endif
}


/**********************************************************************
 * experimental  TAG class
 * ********************************************************************/

enum {INPUT_KIND_DUMMY=1 , INPUT_KIND_INT, INPUT_KIND_DOUBLE, INPUT_KIND_STRING, INPUT_KIND_NONE};
/* INPUT_KIND_NONE is for the internal use */



static int type2inputkind(char *s)
{
      if ( strcasecmp(s,"int")==0) {
              return INPUT_KIND_INT;
      }
      else if (strcasecmp(s,"double")==0 ) {
              return INPUT_KIND_DOUBLE;
      }
      else if (strcasecmp(s,"char*")==0 || strcasecmp(s,"string")==0) {
              return INPUT_KIND_STRING;
      }
      else if (strcasecmp(s,"dummy")==0) {
              return INPUT_KIND_DUMMY;
      }
      return INPUT_KIND_NONE;
}

void tag_construct(TAGCLASS *tag, char *str0, int ntypes, char *typestr[],char *typekind[], int printvalue)
{
  char *c,*s,*str;
  int i;
  int itypes;

  str = strdup(str0);

  s=str;
  c=str;
  i=0;
  while  (c) {
    c=strtok( s, delim);
    if (c) {
      i++;      
    }
    s = NULL;
  }
  tag->ntagorder=i;
  tag->tagorder = ( TAGKIND*)malloc(sizeof( TAGKIND)*i);
  free(str);

  str = strdup(str0);
  s=str;
  c=str;
  i=0;
  while  (c) {
    c=strtok( s, delim);
    if (c) {
      if ( strcasecmp(c, "dummy")==0 ) {
        tag->tagorder[i].i = -1;
        tag->tagorder[i].kind= INPUT_KIND_DUMMY; 
      }
      else {
        for (itypes=0;itypes<ntypes;itypes++) {
          if ( strcasecmp(c, typestr[itypes])==0 ) {
            tag->tagorder[i].i = itypes;
            tag->tagorder[i].kind= type2inputkind(typekind[itypes]); 
            goto FOUND;
          }
        }
        printf("\n\n\nunknown tag =<%s>\n",c);
        printf("at %s\n\n\n",str0);
      }
    FOUND:
      i++;
    }
    s = NULL;
  }
   
  if (printvalue) {
  for (i=0;i<tag->ntagorder;i++) {
    printf("%i(order=%d kind=%d) ",i,tag->tagorder[i].i, tag->tagorder[i].kind);
  }
  printf("\n");
  }
  free(str); 
}

void tag_destruct(TAGCLASS *tag)
{
	    if (tag->tagorder) {free(tag->tagorder);}
	    tag->ntagorder=0;
}


static void tag_tagval_print1(TAGVAL *val)
{
  if (val->success==INPUT_KIND_INT) {
    printf("(%i) ",val->ival);
  }
  else if (val->success==INPUT_KIND_DOUBLE) {
   printf("(%lf) ",val->dval);
  }
  else if (val->success==INPUT_KIND_STRING) {
   printf("(%s) ",val->sval);
  }
  else {
   printf("(NONE) ");
  }
   printf("%d %d %lf <%s>\n",val->success, val->ival,val->dval,val->sval); 
}

void tag_tagval_print(int nval,TAGVAL *val)
{
  int i;
  for (i=0;i<nval;i++){
     tag_tagval_print1(&val[i]);
  }
  printf("\n");
}

int tag_buf2vals(TAGCLASS *tag, char *buf0,int ntagval, TAGVAL *tagval,int printvalue)
{
  char *s,*c;
  int i,po;
  char buf[BUFSIZE];

  for (i=0;i<ntagval;i++) {
    tagval[i].success=0;
    tagval[i].ival=0;
    tagval[i].dval=0;
    if (tagval[i].sval) free( tagval[i].sval );
    tagval[i].sval=NULL;
  } 

  /* use strtok(), which destroys the content of buf */
  if (buf0[strlen(buf0)-1]=='\n') {
    buf0[strlen(buf0)-1]='\0';
  }
  strcpy(buf,buf0);

  c=buf;
  s=buf;
  i=0;
  po=0;
  while (c) {
    if (i>=tag->ntagorder) break;

    c = strtok(s,delim);
    if (c) {
      if ( tag->tagorder[i].kind==INPUT_KIND_DUMMY ) {

      }
      else {
        tagval[ tag->tagorder[i].i ].sval = strdup(c);       
        tagval[ tag->tagorder[i].i ].success=INPUT_KIND_NONE;
        if ( tag->tagorder[i].kind==INPUT_KIND_INT ) {
          if (is_integer(c) )  {
            if (sscanf(c,"%i", &(tagval[ tag->tagorder[i].i ].ival)) ) {
              tagval[ tag->tagorder[i].i ].success=INPUT_KIND_INT;
            }
            else {
              po++;
            }
          } 
          else {
            po++;
          }
        }
        else if ( tag->tagorder[i].kind==INPUT_KIND_DOUBLE ){
          if ( is_double(c) )  {
            if (sscanf(c,"%lf", &(tagval[ tag->tagorder[i].i ].dval))) {
              tagval[ tag->tagorder[i].i ].success=INPUT_KIND_DOUBLE;
            }
            else {
              po++;
            }
          }
          else {
            po++;
          }
                          
        }
        else if ( tag->tagorder[i].kind==INPUT_KIND_STRING ) {
          tagval[ tag->tagorder[i].i ].success=INPUT_KIND_STRING;
        }
      }
      i++;
        
    }
    s = NULL;
  }
  if (po) {
    printf("\n\n\nformat error in \n<%s>\n\n\n", buf0);
  }
  if (printvalue) {
     tag_tagval_print(ntagval,tagval);
  }
  return i;
}

/*
int tag_getsFormat(TAGCLASS *tag,FILE *fp, TAGVAL *tagval)
{
  int n;
  char buf0[BUFSIZE]; 

  fgets(buf0,BUFSIZE,fp);
  n= input_buf2Format(BUFSIZE,buf0, tagval);
  return n;
}

*/

void tag_tagval_init(int n, TAGVAL *val)
{
   int i;
   for (i=0;i<n;i++)  {
      val[i].success=0;
      val[i].sval=NULL;
   }
}

void tag_tagval_destruct(int n , TAGVAL *val)
{
   int i;
   for (i=0;i<n;i++) {
	   if (val[i].sval) free(val[i].sval);
	   val[i].sval=NULL;
   }
}


#if  0
int main()
{
  TAGCLASS tag;
#define NTYPE 8
  const int ntype=NTYPE;
  char *typestr[NTYPE]=  {"x", "y", "z", "typestr","atomid", "moment", "m_theta","m_phi"};
  char *typekind[NTYPE]={"double","double","double","char*","int", "double", "double", "double"};
  int printvalue=0;
  char *buf0;
  int i,n;

  const int ntagval=NTYPE;
  TAGVAL tagval[NTYPE];
  tag_tagval_init(ntagval,tagval);

  tag_construct(&tag, " typestr x y z m_theta m_phi", ntype, typestr,typekind,  printvalue);

  buf0=" H 1.1 2.1 3.1  0.5  1.2 "; 
  n=tag_buf2vals(&tag, buf0,ntagval, tagval,printvalue);
  tag_tagval_print(ntagval,tagval);
  buf0=" H 1.1 2.1 3.1  0.5  1.2 ";
  n=tag_buf2vals(&tag, buf0,ntagval, tagval,printvalue);
  tag_tagval_print(ntagval,tagval);

  tag_tagval_destruct(ntagval,tagval);
  tag_destruct(&tag);
}

#endif



/************************************************
 *  simple iterator library  *
 ************************************************/
typedef struct _IteratorStr {
  struct _IteratorStr *next;
  struct _IteratorStr *branch;
  char *name;
} IteratorStr;

void IteratorStr_init(IteratorStr **top)
{
  *top=NULL;
}

IteratorStr *IteratorStr_next(IteratorStr *top)
{
  if (top) {
    return top->next;
  }
  else {
    return NULL;
  }
}

char *IteratorStr_this(IteratorStr *top)
{
  if (top) {
    return top->name;
  }
  else {
    return NULL;
  }
}

void IteratorStr_add(IteratorStr **top,const char *str)
{
  /* find the last */
  IteratorStr *now;
  if (*top==NULL) {
    *top=(IteratorStr*)malloc(sizeof(IteratorStr));
#ifdef DEBUG
    printf("new %x\n",*top);
#endif
    now=*top;
  }
  else {
    for (now=*top; now->next ; now=( IteratorStr*)now->next) {
    }
    now->next = malloc(sizeof(IteratorStr));
#ifdef DEBUG
    printf("new %x\n",now->next);
#endif
    now =( IteratorStr*)now->next;
  }

  now->next = NULL;
  now->branch=NULL;
  now->name = (char*)strdup(str);
#ifdef DEBUG
  printf("new str %x\n",now->name);
#endif
}


void IteratorStr_push( IteratorStr **top,const char *str)
{
  /* find the last */
  IteratorStr *now;
  IteratorStr *tmp;
  /*   top => top->next => top->next->next 
       tmp=top->next
       new now= top->next
       now->next = tmp
       top => now => tmp => ...
  */

  if (*top==NULL) {
    *top=(IteratorStr*)malloc(sizeof(IteratorStr));
#ifdef DEBUG
    printf("new %x\n",*top);
#endif
    now = *top;
    now->next=NULL;
    now->branch=NULL;
    now->name=strdup(str);
#ifdef DEBUG
    printf("new str %x\n",now->name);
#endif
  }
  else {
    tmp=(IteratorStr*)malloc(sizeof(IteratorStr));
#ifdef DEBUG
    printf("new %x\n",tmp);
#endif
    tmp->name = (*top)->name;
    tmp->next = (*top)->next;
    tmp->branch = (*top)->branch;

    (*top)->next = tmp;
    (*top)->branch=NULL;
    (*top)->name = strdup(str);
#ifdef DEBUG
    printf("new str %x\n",(*top)->name);
#endif
  }
}


int IteratorStr_printAll(IteratorStr *top)
{
  IteratorStr *now;
  int i;
  static int indent=0;
  int tot;
  tot=0;
  for (now=top; now; now=(IteratorStr*)now->next) {
    for (i=0;i<indent*4;i++) {printf(" ");}
    tot++;
    printf("%i %s\n",tot, now->name); 
    if (now->branch) { indent++; tot+=IteratorStr_printAll(( IteratorStr*)now->branch); indent--; }
  }
  return tot;
}

IteratorStr *IteratorStr_find(IteratorStr *top,const char *name)
{
  IteratorStr *now;
  for (now =top; now; now=(IteratorStr*)now->next) {
    if ( strcasecmp(now->name,name) ==0 ) {
      return now; 
    }
  }
  return NULL;
}

void IteratorStr_delete(IteratorStr **top)
{
  IteratorStr *now, *tmp;

  if (*top ==NULL) return; 

  now = *top;
  while ( now ) {
    IteratorStr_delete(&(now->branch));
    if (now->name) { 
#ifdef DEBUG
      printf("free str %x (%s)\n",now->name,now->name);
#endif
      free(now->name);
    }
    tmp = now->next;
#ifdef DEBUG
    printf("free %x\n",now);
#endif
    free(now);
    now = tmp;
  }

}


char *IteratorStr_pop(IteratorStr **top, char *buf)
{
  IteratorStr *tmp;

  if ((*top)->name==NULL) { return NULL; }

  tmp = (*top)->next;
  strcpy(buf,(*top)->name);

  if (tmp==NULL) {
    if ((*top)->name) { 
#ifdef DEBUG
      printf("free str %x\n",(*top)->name);
#endif
      free((*top)->name); 
    }
    if ((*top)->branch) {
      IteratorStr_delete(&((*top)->branch));
    }
    free(*top);
    *top=NULL;
  }
  else {
    if ((*top)->branch) {
      IteratorStr_delete(&((*top)->branch));
    }
    if ((*top)->name) { 
#ifdef DEBUG
      printf("free str %x\n",(*top)->name);
#endif
      free((*top)->name); 
    }

#ifdef DEBUG
    printf("free %x\n",*top);
#endif

    *top = tmp;
  }

}



/************************************************************************/

/*************************************************************************/

#if 0
int main(int argc, char **argv)
{
  IteratorStr *filetop;
  IteratorStr *nametop;
  char buf[100];

  IteratorStr_init(&filetop);

  IteratorStr_add(&filetop,"fileA");

  
  IteratorStr_add(&(filetop->branch),"nameA");
  IteratorStr_add(&(filetop->branch),"nameB");
  IteratorStr_add(&(filetop->branch),"nameC");

  IteratorStr_add(&filetop,"fileB");
  nametop = IteratorStr_find(filetop,"fileB");
  IteratorStr_init(&(nametop->branch));
  IteratorStr_add(&(nametop->branch),"nameD");

  IteratorStr_add(&filetop,"fileC");

  IteratorStr_push(&filetop,"fileD");
  IteratorStr_push(&filetop,"fileE");

  IteratorStr_printAll(filetop);

  printf("=========\n");
  IteratorStr_pop(&filetop, buf);
  IteratorStr_printAll(filetop);
  printf("=========\n");
  IteratorStr_pop(&filetop, buf);
  IteratorStr_printAll(filetop);
  printf("=========\n");
  IteratorStr_pop(&filetop, buf);
  IteratorStr_printAll(filetop);
  printf("=========\n");
  IteratorStr_pop(&filetop, buf);
  IteratorStr_printAll(filetop);


  IteratorStr_delete(&filetop);

}

#endif


/*********************************************************
 * Inputf library
 * use IteratorStr
 *********************************************************/


#define CHECK_TYPE


static int errorlevel=0;
static int printlevel=0;


#define RET_NORMAL  1
#define RET_ERROR  -1
#define RET_NONE    0

#define ERROR_MSG_withStr "\nERROR, key=%s value=(%s)\n"
#define ERROR_MSG         "\nERROR, key=%s\n"
#define KEY_MSG           "keyword=%s value="
#define KEYONLY_MSG       "keyword=%s\n"




static int input_cmpstring(const char *str,  int *ret, int nvals,
			   char **strval,
			   int *ivals)
{
  int i;
  int lenstr,lenstrval;

  /*   for (i=0;i<nvals;i++) {
   *          printf("<%s> <%d>| ",strval[i],ivals[i]);
   *                 }
   *                        printf("\n");
   *                          */

  lenstr=strlen_trim(str);
  for (i=0;i<nvals;i++) {
    lenstrval = strlen_trim(strval[i]);
    if (printlevel>10) {
      printf("%d %d %s %s\n",lenstr,lenstrval,str,strval[i]);
    }
    if (lenstr==lenstrval && strncasecmp(str,strval[i],lenstr)==0 ) {
      if (printlevel>10) {
        printf("<%s> found\n",strval[i]);
      }
      *ret= ivals[i] ;
      if (printlevel>0) {
        printf("%s=%s %d\n",str,strval[i],ivals[i]);
      }
      return 1;
    }
  }
  return 0;
}


static FILE *inputf_find_buf(FILE *fp,const char *key,int size, char *buf, IteratorStr **top)
{
  char key_inbuf[BUFSIZE];
  int keylen,buflen;
  int n;

  
  keylen=strlen_trim(key);

  rewind(fp);
  while ( fgets(buf,size,fp) ) {
    key_inbuf[0]='\0';
    n=sscanf(buf,"%s",key_inbuf); /* can not use strtok, it changes buf */
    if (n<=0) continue;
    buflen = strlen(key_inbuf);
    if (keylen==buflen && strncasecmp(key,key_inbuf,keylen)==0) {
      if (buf[strlen(buf)-1]=='\n') { buf[strlen(buf)-1]=' '; }
      if (top) { IteratorStr_add(top, key); }
      return fp;
    }
  }
  return NULL;
}

FILE  *inputf_find(FILE *fp, const char *key, IteratorStr **top,int printvalue)
{
  int size=BUFSIZE;
  char buf[BUFSIZE];
  FILE *retfp;
  retfp= inputf_find_buf(fp,key,size,buf,top);
  if (retfp) {
    printf(KEYONLY_MSG,key);
  }
  return retfp;
}

int inputf_last(FILE *fp, const char *key0,int printvalue)
{
  char buf[BUFSIZE];
  char inbuf[BUFSIZE];
  char key[BUFSIZE];
  int keylen,buflen;
  int i;

  if (*key0=='<') { 
    strcpy(key,key0+1);
    string_trim(key);
    strcat(key,">");
  }
  else {
    strcpy(key,key0);
  }	
  keylen=strlen_trim(key);

  *inbuf='\0'; /* iniital value */

  /* no rewind */

  fgets(buf,BUFSIZE,fp);
  i=sscanf(buf,"%s",inbuf);

  while (1) {
    /* blank or comments, then read another */
    if ( i!=1 || strlen_trim(buf)==0 ||
	 *inbuf ==cCOMMENT1 || *inbuf==cCOMMENT2 ) {
      fgets(buf,BUFSIZE,fp);
      i=sscanf(buf,"%s",inbuf);
    }
    else {
      break;
    }
  }

  buflen=strlen(inbuf);
  if ( buflen == keylen && strncasecmp(inbuf,key,buflen)==0) {
    printf(KEYONLY_MSG,key);
    return RET_NORMAL;
  }
  else {
    printf(ERROR_MSG,key);
    return RET_ERROR;
  }
}


int inputf_string(FILE *fp, const char *key, char *ret, const char *defval,IteratorStr **top,int printvalue)
{
  int size=BUFSIZE;
  char buf[BUFSIZE];
  char *line;
  char *c;
  int n;
  int returncode;

  if (defval) { strcpy(ret,defval); }

  fp = inputf_find_buf(fp,key,size,buf,top);
  
  if (fp) {
    line = buf;
    c = strtok(line,delim);
    /* c == key */
    line=NULL; 
    c = strtok(line,delim);
    if (c) {
      strcpy(ret,c);
      returncode= RET_NORMAL;
      goto on_return;
    }
    else {
      printf(ERROR_MSG,key);
      errorlevel++;
      returncode= RET_ERROR;
      goto on_return;
    }
  }
  else { /* no keyword */
    if (defval) { returncode= RET_NONE; }
    else { returncode= RET_ERROR; }
  }

 on_return:
  if (returncode!=RET_ERROR) {
    if (printvalue) {
      printf(KEY_MSG,key);
      printf("(%s)\n",ret);
    }
  }
  return returncode;
}

int inputf_stringall( FILE *fp, const char *key, char *ret,const char *defval,IteratorStr **top,int printvalue)
{
  int size=BUFSIZE;
  char buf[BUFSIZE];
  char buf0[BUFSIZE];
  char *c,*s;
  int n;
  int returncode;

  if (defval) { strcpy(ret,defval); }
  fp = inputf_find_buf(fp,key,size,buf0,top);

  if (fp) {

    strcpy(buf,buf0);

    c = strtok(buf,delim);
    /* c == key */
    s = strstr(buf0, c) ;
    strcpy(ret,s+strlen(c)+1);
    string_adjustl(ret);
    string_trim(ret);

    printf("len=%d\n",strlen_trim(ret));
    if ( strlen_trim(ret)==0 ) {
      printf(ERROR_MSG,key);
      errorlevel++;
      returncode= RET_ERROR;
      goto on_return;
    }
    returncode= RET_NORMAL;
    goto on_return;
  }
  else {
    if (defval) {returncode= RET_NONE; }
    else { returncode= RET_ERROR; }
  }

 on_return:

  if (returncode!=RET_ERROR) {
    if (printvalue) {
      printf(KEY_MSG,key);
      printf("(%s)\n",ret);
    }
  }
  return returncode;
}


int inputf_int(FILE *fp, const char *key, int *ret, const int defval, IteratorStr **top,int printvalue)
{
  int size=BUFSIZE;
  char buf[BUFSIZE];
  char *line;
  char *c;
  int n;
  int returncode;

  { *ret=defval; }

  fp = inputf_find_buf(fp,key,size,buf,top);

  if (fp) {
    line = buf;
    c = strtok(line,delim);
    /* c == key */
    line=NULL;
    c = strtok(line,delim);
    if (c) {
      if (is_integer(c)) { sscanf(c,"%i",ret); returncode= RET_NORMAL; goto on_return;}
      else { returncode=RET_ERROR; goto on_return; }
    }
    else {
      returncode=RET_ERROR;
      goto on_return; 

    }
  }
  else { /* no keyword */
    { returncode= RET_NONE; }
  }

 on_return:
  if (returncode==RET_ERROR) {
    printf(ERROR_MSG_withStr,key,c);
    errorlevel++;
  }
  else {
    if (printvalue) {
      printf(KEY_MSG,key);
      printf("(%i)\n",*ret);
    }
  }
  return returncode;

}

int inputf_double(FILE *fp, const char *key, double *ret, const double defval, IteratorStr **top,int printvalue)
{
  int size=BUFSIZE;
  char buf[BUFSIZE];
  char *line;
  char *c;
  int n;
  int returncode;

  { *ret=defval; }

  fp = inputf_find_buf(fp,key,size,buf,top);

  if (fp) {
    line = buf;
    c = strtok(line,delim);
    /* c == key */
    line=NULL;
    c = strtok(line,delim);
    if (c) {
      if (is_double(c)) { sscanf(c,"%lf",ret); returncode= RET_NORMAL; goto on_return;}
      else { returncode=RET_ERROR; goto on_return; }
    }
    else {
      returncode=RET_ERROR;
      goto on_return;
    }
  }
  else { /* no keyword */
    { returncode= RET_NONE; }
  }

 on_return:
  if (returncode==RET_ERROR) {
    printf(ERROR_MSG_withStr,key,c);
    errorlevel++;
  }
  else {
    if (printvalue) {
      printf(KEY_MSG,key);
      printf("(%lf)\n",*ret);
    }
  }
  return returncode;

}

#define SIZE_LOGICAL_DEFAULT 10
static   char *strval[SIZE_LOGICAL_DEFAULT]={"on","yes","true",".true.","ok",
                                             "off","no","false",".false.","ng"};
static  int ival[SIZE_LOGICAL_DEFAULT]={1,1,1,1,1,0,0,0,0,0};

int inputf_logical(FILE *fp, const char *key, int *ret,const  int defval, IteratorStr **top,int printvalue)
{
  int size=BUFSIZE;
  char buf[BUFSIZE];
  char *line;
  char *c;
  int n;
  int returncode;

  { *ret = defval; }

  fp = inputf_find_buf(fp,key,size,buf,top);
  
  if (fp) {
    line = buf;
    c = strtok(line,delim);
    /* c == key */
    line=NULL; 
    c = strtok(line,delim);
    if (c) {
      if ( input_cmpstring(c,  ret, SIZE_LOGICAL_DEFAULT, strval, ival)) {
	returncode= RET_NORMAL;
	goto on_return;
      }
      else {
	returncode=RET_ERROR;
	goto on_return;
      }
    }
    else {
      returncode=RET_ERROR;
      goto on_return; 
    }
  }
  else {
    returncode= RET_NONE;
  }
 on_return:
  if (returncode==RET_ERROR) {
    printf(ERROR_MSG_withStr,key,c);
    errorlevel++;
  }
  else {
    if (printvalue) {
      printf(KEY_MSG,key);
      printf("(%i)\n",*ret);
    }
  }
  return returncode;
}


int inputf_string2int(FILE *fp, const char *key, int *ret, int nvals,
		      char *strval[],  int ivals[], IteratorStr **top,int printvalue)
{
  int size=BUFSIZE;
  char buf[BUFSIZE];
  char *line;
  char *c;
  int n;
  int returncode;

  if (nvals==0) { return RET_ERROR; }
  if (ival==NULL) { return RET_ERROR; }  
  if (strval==NULL) { return RET_ERROR; }

  *ret = ivals[0]; /* default value */ 

  fp = inputf_find_buf(fp,key,size,buf,top);
  
  if (fp) {
    line = buf;
    c = strtok(line,delim);
    /* c == key */
    line=NULL;
    c = strtok(line,delim);
    if (c) {
      if ( input_cmpstring(c,  ret, nvals,strval,ivals) ) { 
	returncode= RET_NORMAL;
	goto on_return;
      }
      else {
	returncode=RET_ERROR;
	goto on_return;
      }
    }
    else {
      returncode=RET_ERROR;
      goto on_return;
    }
  }
  else {
    returncode=RET_NONE;
  }
 on_return:
  if (returncode==RET_ERROR) {
    printf(ERROR_MSG_withStr,key,c);
    errorlevel++;
  }
  else {
    if (printvalue) {
      printf(KEY_MSG,key);
      printf("(%i)\n",*ret);
    }
  }
  return returncode;

}


int inputf_stringv(FILE *fp, const char *key, int nret, char *ret[], const char *defval[], IteratorStr **top, int printvalue)
{
  int ret_code;
  int size=BUFSIZE;
  char buf[BUFSIZE];
  char *line;
  char *c;
  int n;
  int i;
  int returncode;

  if (defval) {  for (i=0;i<nret;i++) strcpy(ret[i],defval[i]); }

  fp = inputf_find_buf(fp,key,size,buf,top);
  
  if (fp) {
    line = buf;
    c = strtok(line,delim);
    /* c == key */
    line=NULL;
    for (i=0;i<nret;i++) {
      c = strtok(line,delim);
      if (c) {
        strcpy(ret[i],c);
      }else { 
	returncode=RET_ERROR;
        goto on_return;
      }
    }
    returncode= nret;
    goto on_return;
  }
  else {
    if (defval) { returncode= RET_NONE; }
    else { returncode= RET_ERROR; }
  }

 on_return:
  if (returncode==RET_ERROR) {
    printf(ERROR_MSG,key);
    errorlevel++;
  }
  else {
    if (printvalue) {
      printf(KEY_MSG,key);
      for (i=0;i<nret;i++) {
	printf(" (%s)",ret[i]);
      }
      printf("\n");
    }
  }
  return returncode;
}


int inputf_intv(FILE *fp, const char *key, int nret, int ret[], const int defval[], IteratorStr **top,int printvalue)
{
  int ret_code;
  int size=BUFSIZE;
  char buf[BUFSIZE];
  char *line;
  char *c;
  int n;
  int i;
  int returncode;

  if (defval) {  for (i=0;i<nret;i++) ret[i]=defval[i]; }

  fp = inputf_find_buf(fp,key,size,buf,top);

  if (fp) {
    line = buf;
    c = strtok(line,delim);
    /* c == key */
    line=NULL;
    for (i=0;i<nret;i++) {
      c = strtok(line,delim);
      if (c) {
        if (is_integer(c)) { sscanf(c,"%i",&ret[i]); }
        else { returncode=RET_ERROR; goto on_return;  }
      } else {
	returncode=RET_ERROR;
        goto on_return;
      }
    }
    returncode=nret;
    goto on_return;
  }
  else {
    if (defval) { returncode= RET_NONE; }
    else { returncode= RET_ERROR; }
  }

 on_return:
  if (returncode==RET_ERROR) {
    printf(ERROR_MSG_withStr,key,c);
    errorlevel++;
  }
  else {
    if (printvalue) {
      printf(KEY_MSG,key);
      for (i=0;i<nret;i++) {
	printf(" (%i)",ret[i]);
      }
      printf("\n");
    }
  }
  return returncode;

}


int inputf_doublev(FILE *fp, const char *key, double nret, double ret[], const double defval[], IteratorStr **top,int printvalue)
{
  int ret_code;
  int size=BUFSIZE;
  char buf[BUFSIZE];
  char *line;
  char *c;
  int n;
  int i;
  int returncode;

  if (defval) {  for (i=0;i<nret;i++) ret[i]=defval[i]; }

  fp = inputf_find_buf(fp,key,size,buf,top);

  if (fp) {
    line = buf;
    c = strtok(line,delim);
    /* c == key */
    line=NULL;
    for (i=0;i<nret;i++) {
      c = strtok(line,delim);
      if (c) {
        if (is_double(c)) { sscanf(c,"%lf",&ret[i]); }
        else { returncode=RET_ERROR; goto on_return;  }
      } else {
	returncode=RET_ERROR;
        goto on_return;
      }
    }
    returncode=nret;
    goto on_return;
  }
  else {
    if (defval) { returncode= RET_NONE; }
    else { returncode= RET_ERROR; }
  }

 on_return:
  if (returncode==RET_ERROR) {
    printf(ERROR_MSG_withStr,key,c);
    errorlevel++;
  }else {
    if (printvalue) {
      printf(KEY_MSG,key);
      for (i=0;i<nret;i++) {
	printf(" (%lf)",ret[i]);
      }
      printf("\n");
    }
  }
  return returncode;

}

int inputf_logicalv(FILE *fp, const char *key, int nret, int ret[], const int defval[], IteratorStr **top,int printvalue)
{
  int ret_code;
  int size=BUFSIZE;
  char buf[BUFSIZE];
  char *line;
  char *c;
  int n;
  int i;
  int returncode;

  if (defval) {  for (i=0;i<nret;i++) ret[i]=defval[i]; }

  fp = inputf_find_buf(fp,key,size,buf,top);

  if (fp) {
    line = buf;
    c = strtok(line,delim);
    /* c == key */
    line=NULL;
    for (i=0;i<nret;i++) {
      c = strtok(line,delim);
      if (c) {
        if ( input_cmpstring(c,  &ret[i], SIZE_LOGICAL_DEFAULT, strval, ival)) {
          sscanf(c,"%lf",&ret[i]); 
        }
        else { returncode=RET_ERROR;goto on_return;  }
      } else {
	returncode=RET_ERROR;
        goto on_return;
      }
    }
    returncode= nret;
    goto on_return;
  }
  else {
    if (defval) { returncode= RET_NONE; }
    else { returncode= RET_ERROR; }
  }

 on_return:
  if (returncode==RET_ERROR) {
    printf(ERROR_MSG_withStr,key,c);
    errorlevel++;
  }else {
    if (printvalue) {
      printf(KEY_MSG,key);
      for (i=0;i<nret;i++) {
	printf(" (%i)",ret[i]);
      }
      printf("\n");
    }
  } 
  return returncode;

}

int inputf_printunknown(FILE *fp, IteratorStr *top)
{
  char buf[BUFSIZE];
  char key_inbuf[BUFSIZE];
  int buflen,keylen;
  IteratorStr *now;
  char *key;
  int found;
  int n;
  char endmark[BUFSIZE];
  int nunknown;

  nunknown=0;
  rewind(fp);

  while ( fgets(buf,BUFSIZE,fp) ) {
    key_inbuf[0]='\0';
    n=sscanf(buf,"%s",key_inbuf); /* can not use strtok, it changes buf */
    if (n<=0) continue;
    buflen = strlen(key_inbuf);
    if (buflen==0) continue;
    /* comment? */
    if (*key_inbuf == cCOMMENT1 ||    *key_inbuf == cCOMMENT2 ) {
#ifdef DEBUG
      printf("skipping comment (%s)\n",key_inbuf);
#endif
      continue;
    }
#ifdef DEBUG
    printf("check name=(%s)\n",key_inbuf);
#endif
    found=0;
    for ( now=top; now; now=IteratorStr_next(now) ) {
      key = IteratorStr_this(now) ;
      keylen = strlen_trim( key );
      if (keylen==buflen && strncasecmp(key,key_inbuf,keylen)==0) {
	found=1;
	break;
      }
    }
    if (found==0) {
      printf("unknown keyword = (%s)\n", key_inbuf);
      nunknown++;
    }

    /* block structure ? */
    if (key_inbuf[0]=='<') {
      /* make end mark */
      strcpy(endmark,&key_inbuf[1]);
      strcat(endmark,">");
      keylen=strlen(endmark);
      while (fgets(buf,BUFSIZE,fp)) {
	key_inbuf[0]='\0';
	n=sscanf(buf,"%s",key_inbuf);
	if (n<=0) continue;
	buflen = strlen(key_inbuf);
	if (buflen==0) continue;
	if (buflen==keylen && strncasecmp(endmark,key_inbuf,buflen)==0) {
	  break;
	}
      }

    }

  }
  return nunknown;
}

int inputf_printdupkey(FILE *fp)
{
  char buf[BUFSIZE];
  char key_inbuf[BUFSIZE];
  int buflen,keylen;
  IteratorStr *top,*now;
  char *key;
  int found;
  int n;
  char endmark[BUFSIZE];
  int ndup;

  top=NULL;
  ndup=0;
  rewind(fp);

  while ( fgets(buf,BUFSIZE,fp) ) {
    key_inbuf[0]='\0';
    n=sscanf(buf,"%s",key_inbuf); /* can not use strtok, it changes buf */
    if (n<=0) continue;
    buflen = strlen(key_inbuf);
    if (buflen==0) continue;
    /* comment? */
    if (*key_inbuf == cCOMMENT1 ||    *key_inbuf == cCOMMENT2 ) {
#ifdef DEBUG
      printf("skipping comment (%s)\n",key_inbuf);
#endif
      continue;
    }
#ifdef DEBUG
    printf("check name=(%s)\n",key_inbuf);
#endif
    found=0;
    for ( now=top; now; now=IteratorStr_next(now) ) {
      key = IteratorStr_this(now) ;
      keylen = strlen_trim( key );
      if (keylen==buflen && strncasecmp(key,key_inbuf,keylen)==0) {
	found=1;
	break;
      }
    }
    if (found==1) {
      printf("duplicate keyword = (%s)\n", key_inbuf);
      ndup++;
    }
    IteratorStr_add(&top,key_inbuf);

    /* block structure ? */
    if (key_inbuf[0]=='<') {
      /* make end mark */
      strcpy(endmark,&key_inbuf[1]);
      strcat(endmark,">");
      keylen=strlen(endmark);
      while (fgets(buf,BUFSIZE,fp)) {
	key_inbuf[0]='\0';
	n=sscanf(buf,"%s",key_inbuf);
	if (n<=0) continue;
	buflen = strlen(key_inbuf);
	if (buflen==0) continue;
	if (buflen==keylen && strncasecmp(endmark,key_inbuf,buflen)==0) {
	  break;
	}
      }

    }

  }

  IteratorStr_delete(&top);
  return ndup;
}


#if 0
int main()
{
  int ival;
  double dval;
  char buf[500]; 
  static int iv[3]={1,2,3};
  static char *sv[3]={"aa","ab","ac"}; 
  int ivval[10];
  double dvval[10];
  static double dv[3]={1.0, 2.0, 3.0}; 
  char *svval[3];
  FILE *fp;
  int i;
  char *key;
  int ret;
  int printvalue=1;

  IteratorStr *top;

  IteratorStr_init(&top);

  svval[0]=malloc(20);
  svval[1]=malloc(20);
  svval[2]=malloc(20); 

  fp = fopen("IN","r");

  key="int2";
  ret=inputf_int(fp,key,&ival,10,&top,printvalue);
  printf("key=%s %i ret=%i\n",key,ival,ret,&top);
  key="double";
  ret=inputf_double(fp,key,&dval,1.0,&top,printvalue);
  printf("key=%s %le ret=%i\n",key,dval,ret);
  key="string";
  ret=inputf_string(fp,key,buf,"default",&top,printvalue);
  printf("key=%s <%s> ret=%i\n",key,buf,ret);
  key="logical";
  ret=inputf_logical(fp,key,&ival,1,&top,printvalue);
  printf("key=%s %i ret=%i\n",key,ival,ret);
  key="str2int";
  ret=inputf_string2int(fp,key,&ival, 3, sv , iv,&top,printvalue); 
  printf("key=%s %i ret=%i\n",key,ival,ret); 


  { int n;
  key="<block";
  fp = inputf_find(fp,key,&top,printvalue);
  n=3;
  for (i=0;i<n;i++) {
    fgets(buf,BUFSIZE,fp);
    printf("%s",buf);
  }
  i = inputf_last(fp,key,printvalue);
  printf("ret=%i\n",i);
  }


  printf("--------keywords------------\n");
  i=IteratorStr_printAll(top);
  printf("----------------------------\n");

  i=inputf_printunknown(fp,top);
  printf("unknown=%i\n",i);

  fclose(fp);
}

#endif



/****************************************************************
 * Input library 
 * use Inputf
 * use IteratorStr
 ****************************************************************/


static IteratorStr *filelisttop=NULL; 
static IteratorStr *thisfile_namelist=NULL;

static int printvalue=0;

#undef DEBUG


#define FILE_LOCATION_MSG "       <-------- in file %s\n\n"
#define OPENFAILER_MSG   "ERROR: (%s) failed to open a file, %s\n"

int input_printflag(int f)
{
  printvalue=f;
}

int input_open(const char *fname)
{
  FILE *fp;
  const char *this="input_open";

  if (fname) {
    IteratorStr_push(&filelisttop,fname);
    thisfile_namelist = IteratorStr_find(filelisttop,fname);
  }

  fp= fopen(fname,"r");
  if (fp==NULL) {
    printf(OPENFAILER_MSG,this,fname); 
    return RET_ERROR;
  }

  fclose(fp);
  return RET_NORMAL;
}

int input_close()
{
  char buf[BUFSIZE];
  char *s;
  IteratorStr_pop(&filelisttop,buf);
  thisfile_namelist = filelisttop;
  printf("close file=(%s)\n",buf);
  s = IteratorStr_this(thisfile_namelist);
  if (s){
    printf("next file=(%s)\n",s);
  }
}

FILE *input_find(const char *key)
{
  char *fname;
  char *this="input_find";
  FILE *fp;
  fname = IteratorStr_this(thisfile_namelist);
  fp= fopen(fname,"r");
  if (fp==NULL) {
    printf(OPENFAILER_MSG,this,fname);
    return NULL;
  }

  fp = inputf_find(fp,key,&(thisfile_namelist->branch),printvalue);
  return fp; 
}

int input_last(FILE *fp,const char *key)
{

  char *fname;
  char *this="input_last";
  int returncode;

  if (fp==NULL) {
    returncode=RET_ERROR;
  }
  else {
    returncode=inputf_last(fp,key,printvalue);
  }

  return returncode;
}


int input_string(const char *key, char *ret, const char *defval)
{
  int retcode;
  char *fname;
  FILE *fp;
  char *this="input_string";

  fname = IteratorStr_this(thisfile_namelist);
#ifdef DEBUG
  printf("fname=(%s)\n",fname);
#endif
  fp= fopen(fname,"r");
  if (fp==NULL) {
    printf(OPENFAILER_MSG,this,fname);
    return RET_ERROR;
  }
  retcode = inputf_string(fp,key,ret,defval,&(thisfile_namelist->branch),printvalue);
  if (retcode==RET_ERROR) {
    printf(FILE_LOCATION_MSG,IteratorStr_this(thisfile_namelist));
  }

  fclose(fp);
  return retcode;
}

int input_stringall(const char *key, char *ret, const char *defval)
{
  int retcode;
  char *fname;
  FILE *fp;
  char *this="input_stringall";

  fname = IteratorStr_this(thisfile_namelist);
#ifdef DEBUG
  printf("fname=(%s)\n",fname);
#endif
  fp= fopen(fname,"r");
  if (fp==NULL) {
    printf(OPENFAILER_MSG,this,fname);
    return RET_ERROR;
  }
  retcode = inputf_stringall(fp,key,ret,defval,&(thisfile_namelist->branch),printvalue);
  if (retcode==RET_ERROR) {
    printf(FILE_LOCATION_MSG,IteratorStr_this(thisfile_namelist));
  }
  fclose(fp);
  return retcode;
}



int input_int(const char *key, int *ret, const int defval)
{
  int retcode;
  char *fname;
  FILE *fp;
  char *this="input_int";

  fname = IteratorStr_this(thisfile_namelist);
#ifdef DEBUG
  printf("fname=(%s)\n",fname);
#endif
  fp= fopen(fname,"r");
  if (fp==NULL) {
    printf(OPENFAILER_MSG,this,fname);
    return RET_ERROR;
  }
  retcode = inputf_int(fp,key,ret,defval,&(thisfile_namelist->branch),printvalue); 
  if (retcode==RET_ERROR) {
    printf(FILE_LOCATION_MSG,IteratorStr_this(thisfile_namelist));
  }

  fclose(fp);
  return retcode;
}

int input_double(const char *key, double *ret, const double defval)
{
  int retcode;
  char *fname;
  FILE *fp;
  char *this="input_double";

  fname = IteratorStr_this(thisfile_namelist);
#ifdef DEBUG
  printf("fname=(%s)\n",fname);
#endif
  fp= fopen(fname,"r");
  if (fp==NULL) {
    printf(OPENFAILER_MSG,this,fname);
    return RET_ERROR;
  }
  retcode = inputf_double(fp,key,ret,defval,&(thisfile_namelist->branch),printvalue);
  if (retcode==RET_ERROR) {
    printf(FILE_LOCATION_MSG,IteratorStr_this(thisfile_namelist));
  }

  fclose(fp);
  return retcode;
}

int input_logical(const char *key, int *ret, const int defval)
{
  int retcode;
  char *fname;
  FILE *fp;
  char *this="input_logical";

  fname = IteratorStr_this(thisfile_namelist);
#ifdef DEBUG
  printf("fname=(%s)\n",fname);
#endif
  fp= fopen(fname,"r");
  if (fp==NULL) {
    printf(OPENFAILER_MSG,this,fname);
    return RET_ERROR;
  }
  retcode = inputf_logical(fp,key,ret,defval,&(thisfile_namelist->branch),printvalue);
  if (retcode==RET_ERROR) {
    printf(FILE_LOCATION_MSG,IteratorStr_this(thisfile_namelist));
  }

  fclose(fp);
  return retcode;
}

int input_string2int(const char *key, int *ret, int nvals,
                     char **strval,  int *ivals)
{
  int retcode;
  char *fname;
  FILE *fp;
  char *this="input_string2int";

  fname = IteratorStr_this(thisfile_namelist);
#ifdef DEBUG
  printf("fname=(%s)\n",fname);
#endif
  fp= fopen(fname,"r");
  if (fp==NULL) {
    printf(OPENFAILER_MSG,this,fname);
    return RET_ERROR;
  }
  retcode = inputf_string2int(fp,key,ret,nvals,
			      strval,ivals, &(thisfile_namelist->branch),printvalue);
  if (retcode==RET_ERROR) {
    printf(FILE_LOCATION_MSG,IteratorStr_this(thisfile_namelist));
  }

  fclose(fp);
  return retcode;

}

int input_stringv(const char *key, int n, char *ret[], const char *defval[])
{
  int retcode;
  char *fname;
  FILE *fp;
  char *this="input_stringv";

  fname = IteratorStr_this(thisfile_namelist);
#ifdef DEBUG
  printf("fname=(%s)\n",fname);
#endif
  fp= fopen(fname,"r");
  if (fp==NULL) {
    printf(OPENFAILER_MSG,this,fname);
    return RET_ERROR;
  }
  retcode = inputf_stringv(fp,key,n,ret,defval,&(thisfile_namelist->branch),printvalue);
  if (retcode==RET_ERROR) {
    printf(FILE_LOCATION_MSG,IteratorStr_this(thisfile_namelist));
  }

  fclose(fp);
  return retcode;
}


int input_intv(const char *key, int n, int *ret, const int *defval)
{
  int retcode;
  char *fname;
  FILE *fp;
  char *this="input_intv";

  fname = IteratorStr_this(thisfile_namelist);
#ifdef DEBUG
  printf("fname=(%s)\n",fname);
#endif
  fp= fopen(fname,"r");
  if (fp==NULL) {
    printf(OPENFAILER_MSG,this,fname);
    return RET_ERROR;
  }
  retcode = inputf_intv(fp,key,n,ret,defval,&(thisfile_namelist->branch),printvalue);
  if (retcode==RET_ERROR) {
    printf(FILE_LOCATION_MSG,IteratorStr_this(thisfile_namelist));
  }

  fclose(fp);
  return retcode;
}

int input_doublev(const char *key, double n, double *ret, const double *defval)
{
  int retcode;
  char *fname;
  FILE *fp;
  char *this="input_doublev";

  fname = IteratorStr_this(thisfile_namelist);
#ifdef DEBUG
  printf("fname=(%s)\n",fname);
#endif
  fp= fopen(fname,"r");
  if (fp==NULL) {
    printf(OPENFAILER_MSG,this,fname);
    return RET_ERROR;
  }
  retcode = inputf_doublev(fp,key,n,ret,defval,&(thisfile_namelist->branch),printvalue);
  if (retcode==RET_ERROR) {
    printf(FILE_LOCATION_MSG,IteratorStr_this(thisfile_namelist));
  }

  fclose(fp);
  return retcode;
}


int input_logicalv(const char *key, int n, int *ret, const int *defval)
{
  int retcode;
  char *fname;
  FILE *fp;
  char *this="input_logicalv";

  fname = IteratorStr_this(thisfile_namelist);
#ifdef DEBUG
  printf("fname=(%s)\n",fname);
#endif
  fp= fopen(fname,"r");
  if (fp==NULL) {
    printf(OPENFAILER_MSG,this,fname);
    return RET_ERROR;
  }
  retcode = inputf_logicalv(fp,key,n,ret,defval,&(thisfile_namelist->branch),printvalue);
  if (retcode==RET_ERROR) {
    printf(FILE_LOCATION_MSG,IteratorStr_this(thisfile_namelist));
  }

  fclose(fp);
  return retcode;
}


int input_printkeywords()
{
  int retcode;
  char *s;
  s = IteratorStr_this(thisfile_namelist);
  printf("----------- keywords defined (file = %s) ------------\n",s);
  retcode= IteratorStr_printAll((thisfile_namelist->branch));
  return retcode;
}

int input_printunknown(void)
{
  int retcode;
  char *fname;
  FILE *fp;
  char *this="input_printunknown";

  fname = IteratorStr_this(thisfile_namelist);
  fp= fopen(fname,"r");
  if (fp==NULL) {
    printf(OPENFAILER_MSG,this,fname);
    return RET_ERROR;
  }

  printf("------------ unknown keywords (file = %s) ------------\n",fname);
  retcode= inputf_printunknown(fp,(thisfile_namelist->branch));
  return retcode;
}

int input_printdupkey(void)
{
  int retcode;
  char *fname;
  FILE *fp;
  char *this="input_printdupkey";

  fname = IteratorStr_this(thisfile_namelist);
  fp= fopen(fname,"r");
  if (fp==NULL) {
    printf(OPENFAILER_MSG,this,fname);
    return RET_ERROR;
  }

  printf("------------ duplicate keywords (file = %s) ------------\n",fname);
  retcode= inputf_printdupkey(fp);
  return retcode;
}



#if 0
int main()
{
  int retcode;	 
  int i;
  char *key;
  double d;
  int nvals=3;
  static int iv[3]={1,2,3};
  static char *sv[3]={"aa","ab","ac"}; 
  static int ivdef[3]={10,11,12};
  static double dv[10];
  char str[100];
  char *strv[10];
  FILE *fp;

  for (i=0;i<10;i++) strv[i]=(char*)malloc(sizeof(char)*100); 


  input_printflag(1);

  input_open("IN");
  key="int";
  retcode=input_int(key,&i,100);
  //  printf("%s ret=%i val=%i\n",key,retcode,i);

  key="double";
  retcode=input_double(key,&d,100.0);
  //  printf("%s ret=%i val=%lf\n",key,retcode,d);

  key="logical";
  retcode=input_logical(key,&i,0);
  //  printf("%s ret=%i val=%i\n",key,retcode,i);

  key="string";
  retcode=input_string(key,str,"test");
  //  printf("%s ret=%i val=(%s)\n",key,retcode,str);

  key="str2int";
  retcode=input_string2int(key,&i,nvals,sv,iv);
  //  printf("%s ret=%i val=(%i)\n",key,retcode,i);

  {
    input_open("IN2");
    key="int";
    retcode=input_int(key,&i,100);
    //    printf("%s ret=%i val=%i\n",key,retcode,i);
    i=input_printkeywords();
    printf("n=%i\n",i);

    i=input_printunknown();
    printf("n=%i\n",i);

    input_close();

  }


  key="intv";
  retcode=input_intv(key,3,iv,ivdef);
  //  printf("%s ret=%i val=%i %i %i\n",key,retcode,iv[0],iv[1],iv[2]);

  key="doublev";
  retcode=input_doublev(key,3,dv,NULL);
  //  printf("%s ret=%i val=%lf %lf %lf\n",key,retcode,dv[0],dv[1],dv[2]);

  key="stringv";
  retcode=input_stringv(key,3,strv,NULL);
  //  printf("%s ret=%i val=(%s) (%s) (%s)\n",key,retcode,strv[0],strv[1],strv[2]);
  
  key="stringv";
  retcode=input_stringall(key,str,NULL);
  //  printf("%s ret=%i val=(%s)\n",key,retcode,str);
  //

  { char buf[BUFSIZE];
  key="<block";
  fp=input_find(key);
  if (fp) {
    for (i=0;i<3;i++) {
      fgets(buf,BUFSIZE,fp);
      printf("%s",buf);
    }
    retcode=input_last(fp,key);
  }
  }

  i=input_printkeywords();
  printf("n=%i\n",i);

  i=input_printdupkey();
  printf("n=%i\n",i);

  i=input_printunknown();
  printf("n=%i\n",i);

  input_close();

}

#endif
